/*
Arcade Physics
	by Lach

this isn't very polished, sorry
redirection doesn't seem to consistently work with all walls, and it doesn't work with solid midtextures
https://cdn.discordapp.com/attachments/705529445313806477/941691518396272650/srb20215.gif
might be improved by MobjMoveBlocked parameters in 2.2.10
ceiling running should work in reverse gravity but I don't really know what happens in gravity changing sectors
you have been warned! C:<

v1
*/

if ArcadePhysics
	return
end

rawset(_G, "ArcadePhysics", true)

// vars

local active = false

// constants

local ORIG_FRICTION = 29*FRACUNIT/32
local CEILING_SPEED_THRESHOLD = 16*FRACUNIT
local WALL_ANGLE_THRESHOLD = ANG15

// functions

local function CreateVarsTable()
	return {
		ceilingRun = false,
		redirectAngle = 0,
		redirectSpeed = 0
	}
end

local function FixedSquare(n)
	return FixedMul(n, n)
end

local function FlipGravity(mo)
	mo.flags2 = $ ^^ MF2_OBJECTFLIP
	mo.eflags = $ ^^ MFE_VERTICALFLIP
end

local function FV3_Rotate(rotVecX, rotVecY, rotVecZ, axisVecX, axisVecY, axisVecZ, angle)
	local ux, uy, uz = FixedMul(axisVecX, rotVecX), FixedMul(axisVecX, rotVecY), FixedMul(axisVecX, rotVecZ)
	local vx, vy, vz = FixedMul(axisVecY, rotVecX), FixedMul(axisVecY, rotVecY), FixedMul(axisVecY, rotVecZ)
	local wx, wy, wz = FixedMul(axisVecZ, rotVecX), FixedMul(axisVecZ, rotVecY), FixedMul(axisVecZ, rotVecZ)
	local sa, ca = sin(angle), cos(angle)
	local ua = ux + vy + wz
	local ax, ay, az = FixedMul(axisVecX, ua), FixedMul(axisVecY, ua), FixedMul(axisVecZ, ua)
	local xs, ys, zs = FixedSquare(axisVecX), FixedSquare(axisVecY), FixedSquare(axisVecZ)
	local bx, by, bz = FixedMul(rotVecX, ys + zs), FixedMul(rotVecY, xs + zs), FixedMul(rotVecZ, xs + ys)
	local cx, cy, cz = FixedMul(axisVecX, vy + wz), FixedMul(axisVecY, ux + wz), FixedMul(axisVecZ, ux + vy)
	local dx, dy, dz = FixedMul(bx - cx, ca), FixedMul(by - cy, ca), FixedMul(bz - cz, ca)
	local ex, ey, ez = FixedMul(vz - wy, sa), FixedMul(wx - uz, sa), FixedMul(uy - vx, sa)
	
	return ax + dx + ex, ay + dy + ey, az + dz + ez
end

local function GetSlopeLandingMomentum(player, slope)
	local mo = player.mo
	
	if not slope
	or slope.flags & SL_NOPHYSICS
	or slope.zangle == 0
		return mo.momx, mo.momy
	end
	
	return FV3_Rotate(mo.momx, mo.momy, 2 * mo.momz, -slope.d.y, slope.d.x, 0, InvAngle(slope.zangle))
end

local function GetFloorSlope(mo, flip)
	if not (mo.eflags & MFE_VERTICALFLIP) == not flip
		local fof = mo.floorrover
		if fof
			return fof.t_slope
		end
		return mo.subsector.sector.f_slope
	else
		local fof = mo.ceilingrover
		if fof
			return fof.b_slope
		end
		return mo.subsector.sector.c_slope
	end
end

local function MobjWillTouchFloor(mo, flip)
	if not (mo.eflags & MFE_VERTICALFLIP) == not flip
		return mo.z + mo.momz < mo.floorz
	else
		return mo.z + mo.height + mo.momz > mo.ceilingz
	end
end

local function GetBlockingSide(player, line) // doesn't account for solid midtextures, no way to get their height
	if line.frontsector == line.backsector
		return
	end

	local mo = player.mo

	if mo.flags & MF_NOCLIP
		return
	end

	local sector
	local side = P_PointOnLineSide(mo.x - mo.momx, mo.y - mo.momy, line)

	if side == 1
		sector = line.frontsector
	else
		sector = line.backsector
	end
	
	if sector == nil
		return 1 - side
	end

	local floor, ceiling
	local x, y = mo.x + mo.momx, mo.y + mo.momy
	local ramp

	if P_PlayerTouchingSectorSpecial(player, 1, 13)
		ramp = FixedMul(2*MAXSTEPMOVE, mo.scale)
	elseif P_PlayerTouchingSectorSpecial(player, 1, 14)
		ramp = 0
	else
		ramp = FixedMul(MAXSTEPMOVE, mo.scale)
	end

	if sector.f_slope
		floor = P_GetZAt(sector.f_slope, x, y)
	else
		floor = sector.floorheight
	end

	if sector.c_slope
		ceiling = P_GetZAt(sector.c_slope, x, y)
	else
		ceiling = sector.ceilingheight
	end

	if floor - ramp > mo.z or ceiling + ramp < mo.z + mo.height
		return 1 - side
	end

	for fof in sector.ffloors()
		if (fof.flags & (FF_EXISTS|FF_BLOCKPLAYER)) ~= (FF_EXISTS|FF_BLOCKPLAYER)
			continue
		end

		if fof.t_slope
			floor = P_GetZAt(fof.t_slope, x, y)
		else
			floor = fof.topheight
		end

		if fof.b_slope
			ceiling = P_GetZAt(fof.b_slope, x, y)
		else
			ceiling = fof.bottomheight
		end

		if floor - ramp > mo.z and ceiling + ramp < mo.z + mo.height
			return 1 - side
		end
	end
	return
end

addHook("PlayerThink", function(player)
	local physics = player.arcadePhysics

	if not physics
		return
	end
	
	local mo = player.mo
	
	if not mo
		physics.ceilingRun = false
		return
	end
	
	if (mo.flags & MF_NOCLIPHEIGHT)
	or player.powers[pw_carry] ~= CR_NONE
		if physics.ceilingRun
			FlipGravity(mo)
			physics.ceilingRun = false
		end
		return
	end

	if P_IsObjectOnGround(mo)
		if not physics.ceilingRun
			return
		end
		
		if FixedHypot(mo.momx - player.cmomx, mo.momy - player.cmomy) < FixedMul(CEILING_SPEED_THRESHOLD, mo.scale)
			FlipGravity(mo)
			physics.ceilingRun = false
		end
		return
	end

	if MobjWillTouchFloor(mo, true)
		local speed = FixedHypot(GetSlopeLandingMomentum(player, GetFloorSlope(mo, true)))
		local threshold = FixedMul(CEILING_SPEED_THRESHOLD, mo.scale)
		
		if speed >= threshold
			FlipGravity(mo)
			physics.ceilingRun = true
		end
	elseif physics.ceilingRun
		FlipGravity(mo)
		physics.ceilingRun = false
	end
end)

addHook("ThinkFrame", do
	if not active
		return
	end
	
	for player in players.iterate
		local mo = player.mo
		
		if not mo
			continue
		end
		
		local physics = player.arcadePhysics
		
		if physics.redirectSpeed
			P_InstaThrust(mo, physics.redirectAngle, physics.redirectSpeed)
			physics.redirectSpeed = 0
		end
	end
end)

addHook("MobjLineCollide", function(mo, line)
	if not active
		return
	end
	
	local player = mo.player

	if not player
	or not (player.pflags & PF_SPINNING)
		return
	end

	local v1, v2 = line.v1, line.v2
	local lineAngle = R_PointToAngle2(v1.x, v1.y, v2.x, v2.y)
	local momAngle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	local angle = abs(momAngle - lineAngle)

	if angle > ANGLE_90 - WALL_ANGLE_THRESHOLD and angle < ANGLE_90 + WALL_ANGLE_THRESHOLD
		return
	end

	local side = GetBlockingSide(player, line)

	if side == nil
		return
	end

	if angle > ANGLE_90
		lineAngle = $ + ANGLE_180
	end
	
	local physics = player.arcadePhysics

	physics.redirectAngle = lineAngle
	physics.redirectSpeed = FixedHypot(mo.momx, mo.momy)
end, MT_PLAYER)

addHook("PlayerSpawn", function(player)
	player.arcadePhysics = active and CreateVarsTable() or nil
end)

addHook("MapChange", function(map)
	map = mapheaderinfo[$]
	active = not not (map and map.arcadephysics)
end)

addHook("NetVars", function(net)
	active = net(active)
end)