/*
Chrispy's Beach Ball
	by Lach

v1
*/

local CA = ChrispyAssets

if CA.BeachBall
	return
end

local BB = {}

CA.BeachBall = BB

CA.SafeFreeslot("SPR_BEBA", "SPR_BEB2", "S_CHRISPY_BEACHBALL", "MT_CHRISPY_BEACHBALL")

states[S_CHRISPY_BEACHBALL] = {
	sprite = SPR_BEBA,
	frame = X,
	tics = -1
}

mobjinfo[MT_CHRISPY_BEACHBALL] = {
	--$Name Beach Ball
	--$Sprite BEBAX0
	--$Category Chrispy Pixels
	doomednum = 834,
	spawnstate = S_CHRISPY_BEACHBALL,
	spawnhealth = 1000,
	radius = 16*FRACUNIT,
	height = 32*FRACUNIT,
	flags = MF_SPECIAL|MF_BOUNCE,
	activesound = sfx_s3k44
}

BB.FRICTION = (CA.ORIG_FRICTION + FRACUNIT) / 2
BB.MOMENTUM_SCALE = 4*FRACUNIT/3
BB.BOUNCE_FACTOR = 3*FRACUNIT/5
BB.MIN_BOUNCE = 2*FRACUNIT
BB.BUOYANCY = 4*FRACUNIT
BB.DESPAWN_TIME = 30*TICRATE

BB.animations = {}
BB.inLevel = false

function BB.AddAnimation(sprite, startFrame, endFrame, axis, addFlipped)
	startFrame = R_Char2Frame($)
	endFrame = R_Char2Frame($)
	table.insert(BB.animations, {
		sprite = sprite,
		startFrame = startFrame,
		endFrame = endFrame,
		axis = axis
	})
	if addFlipped
		table.insert(BB.animations, {
			sprite = sprite,
			startFrame = startFrame|FF_HORIZONTALFLIP,
			endFrame = endFrame|FF_HORIZONTALFLIP,
			axis = axis
		})
	end
end

BB.AddAnimation(SPR_BEBA, 'A', 'L', 1, true)
BB.AddAnimation(SPR_BEBA, 'M', 'X', 2)
BB.AddAnimation(SPR_BEBA, 'Y', '9', 2)
BB.AddAnimation(SPR_BEBA, 'a', 'l', 2)
BB.AddAnimation(SPR_BEB2, 'A', 'L', 3)
BB.AddAnimation(SPR_BEB2, 'M', 'X', 4, true)

function BB.UpdateAnimationSpeed(ball)
	ball.extravalue2 = FixedDiv(FixedHypot(FixedHypot(ball.momx, ball.momy), ball.momz), ball.scale)
end

function BB.SetAnimation(ball, index)
	local anim = BB.animations[index]
	ball.extravalue1 = index - 1
	ball.sprite = anim.sprite
	ball.frame = anim.startFrame
end

function BB.ChangeAnimation(ball)
	local currentAnim = BB.animations[ball.extravalue1 + 1]
	local validAnimations = 0
	local newIndex = 0
	local rand
	
	for _, anim in ipairs(BB.animations)
		if currentAnim.axis ~= anim.axis
			validAnimations = $ + 1
		end
	end
	
	rand = P_RandomKey(validAnimations) + 1
	
	while rand > 0
		rand = $ - 1
		newIndex = $ + 1
		while BB.animations[newIndex].axis == currentAnim.axis
			newIndex = $ + 1
		end
	end
	
	BB.SetAnimation(ball, newIndex)
	BB.UpdateAnimationSpeed(ball)
end

function BB.CollidePlayer(ball, mo)
	// don't allow the ball to touch a player continuously
	local touched = ball.movecount > 0
	ball.movecount = 2
	
	if touched
		return
	end
	
	// get speeds
	local player = mo.player
	local momx = mo.momx - player.cmomx
	local momy = mo.momy - player.cmomy
	local speed = FixedHypot(momx, momy)
	local ballSpeed = FixedHypot(ball.momx, ball.momy)
	
	// set ball's momentum
	if ballSpeed < speed // player kicked ball
		ball.momx = FixedMul(momx, BB.MOMENTUM_SCALE)
		ball.momy = FixedMul(momy, BB.MOMENTUM_SCALE)
		
		ball.momz = P_MobjFlip(mo) * FixedMul(speed, BB.MOMENTUM_SCALE) / 2
	elseif speed == 0 // ball hit stationary player
		P_InstaThrust(ball, R_PointToAngle2(mo.x, mo.y, ball.x, ball.y), ballSpeed)
		
		if abs(ball.momz) > FixedMul(BB.MIN_BOUNCE, ball.scale)
			ball.momz = -FixedMul(ball.momz, BB.BOUNCE_FACTOR)
		end
	else // rebound!
		local ballDir = R_PointToAngle2(0, 0, ball.momx, ball.momy)
		local playerDir = R_PointToAngle2(0, 0, momx, momy)
		
		ballSpeed = FixedMul(max($, speed), BB.MOMENTUM_SCALE)
		
		P_InstaThrust(ball, ballDir + (playerDir - ballDir)/2, ballSpeed)
		ball.momz = P_MobjFlip(mo) * ballSpeed / 2
	end
	
	// boing!
	S_StartSound(ball, ball.info.activesound)
	BB.ChangeAnimation(ball)
	
	// set the despawn fuse
	ball.flags2 = $ & ~MF2_DONTDRAW
	ball.fuse = BB.DESPAWN_TIME
	ball.target = mo
end

function BB.CollideSpring(ball, spring)
	if ball.eflags & MFE_SPRUNG
		return
	end
	
	if spring.spling
		if not CA.Spling.DoSpling(spring, ball)
			return
		end
	else
		if not P_DoSpring(spring, ball)
			return
		end
	end
	
	S_StartSound(ball, ball.info.activesound)
	BB.ChangeAnimation(ball)
end

function BB.CollideSteam(ball, steam)
	if steam.state ~= S_STEAM1
		return
	end
	
	local height = BB.GetHeightAboveFan(ball, steam)
	if height == nil
	or height > 16*steam.scale
		return
	end
	
	local flip = P_MobjFlip(steam)
	local speed = steam.info.mass
	
	ball.eflags = $ | MFE_SPRUNG
	ball.momz = flip * FixedMul(speed, FixedSqrt(FixedMul(steam.scale, ball.scale)))
	BB.ChangeAnimation(ball)
end

function BB.CollideFan(ball, fan)
	local height = BB.GetHeightAboveFan(ball, fan)
	if height == nil
	or height > fan.health << FRACBITS
		return
	end
	
	BB.DoFan(FixedMul(fan.info.mass, fan.scale), P_MobjFlip(fan))
end

function BB.DoFan(ball, speed, flip)
	if flip * ball.momz >= speed
		return
	end
	
	ball.momz = $ + flip * speed / 4
	
	if flip * ball.momz > speed
		ball.momz = flip * speed
	end
end

function BB.GetHeightAboveFan(ball, fan)
	if fan.eflags & MFE_VERTICALFLIP
		if ball.z > fan.z + fan.height
			return
		end
		return (fan.z + fan.height) - (ball.z + ball.height)
	else
		if ball.z + ball.height < fan.z
			return
		end
		return ball.z - fan.z
	end
end

function BB.CollideDustDevil(ball, devil)
	local thresh = devil.scale * 20
	local pos = ball.z - devil.z
	
	if devil.height - pos <= thresh
		ball.z = devil.z + devil.height
		ball.momz = thresh
		S_StartSound(ball, sfx_wdjump)
		BB.ChangeAnimation(ball)
		return
	end
	
	local thrust = max(FixedDiv(pos, devil.height) * 20, 8 * devil.scale)
	local fa = R_PointToAngle2(ball.x, ball.y, devil.x, devil.y)
	local c = cos(fa)
	local s = sin(fa)
	
	local dist = FixedHypot(ball.x - devil.x, ball.y - devil.y)
	local drag = FixedHypot(ball.momx, ball.momy)
	local x, y
	
	if ball.movecount == 0
		A_PlayActiveSound(devil)
	end
	ball.movecount = 2
	
	if dist > drag
		x = ball.x + FixedMul(c, drag)
		y = ball.y + FixedMul(s, drag)
	else
		x = devil.x
		y = devil.y
	end
	P_TryMove(ball, x - ball.momx, y - ball.momy, true)
	ball.momz = thrust
end

function BB.Collide(ball, mo)
	local type = mo.type
	
	// fans
	if type == MT_FAN
		return BB.CollideFan(ball, mo)
	end
	
	// steam jets
	if type == MT_STEAM
		return BB.CollideSteam(ball, mo)
	end
	
	// everything else uses height checks
	if ball.z > mo.z + mo.height
	or mo.z > ball.z + ball.height
		return
	end
	
	// players
	if mo.player
		return BB.CollidePlayer(ball, mo)
	end
	
	// springs
	if mo.flags & MF_SPRING
		return BB.CollideSpring(ball, mo)
	end
	
	// bould shield hitbox
	if SH_BOULDER
	and type == MT_BOULDER_COLLIDER
		local target = mo.target
		if target ~= nil
		and target.player
			return BB.CollidePlayer(ball, target)
		end
		return
	end
	
	// rollout rock
	if type == MT_ROLLOUTROCK
		local tracer = mo.tracer
		if tracer ~= nil
		and tracer.player
			return BB.CollidePlayer(ball, tracer)
		end
		return
	end
end

function BB.DustDevilSearch(devil, ball)
	if ball.type ~= MT_CHRISPY_BEACHBALL
		return
	end
	
	if ball.z > devil.z + devil.height
	or devil.z > ball.z + ball.height
	or abs(ball.x - devil.x) > devil.radius
	or abs(ball.y - devil.y) > devil.radius
		return
	end
	BB.CollideDustDevil(ball, devil)
end

function BB.DoFriction(ball)
	local slope = ball.standingslope
	
	if slope ~= nil
	and abs(slope.zdelta) > FRACUNIT / 4
		if CA.ButteredSlope(ball)
			ball.friction = FRACUNIT
			return
		end
	end
	
	if ball.friction < FRACUNIT
		ball.friction = min(FixedMul(FixedDiv($, CA.ORIG_FRICTION), BB.FRICTION), FRACUNIT)
	end
end

addHook("MobjCollide", BB.Collide, MT_CHRISPY_BEACHBALL)
addHook("MobjMoveCollide", BB.Collide, MT_CHRISPY_BEACHBALL)

addHook("TouchSpecial", do return true end, MT_CHRISPY_BEACHBALL) // MF_SPECIAL needed for collisions

addHook("MobjSpawn", function(ball)
	ball.shadowscale = FRACUNIT
	BB.SetAnimation(ball, P_RandomKey(#BB.animations) + 1)
	BB.inLevel = true
end, MT_CHRISPY_BEACHBALL)

addHook("MobjMoveBlocked", function(ball)
	BB.ChangeAnimation(ball)
end, MT_CHRISPY_BEACHBALL)

addHook("MobjThinker", function(ball)
	// slide down slopes
	BB.DoFriction(ball)
	
	// check whether the ball touched the floor or ceiling
	if abs(ball.movefactor) > FixedMul(BB.MIN_BOUNCE, ball.scale)
		local bounce = 0
		
		// floor bounce
		if (ball.z <= ball.floorz)
			bounce = 1
		// ceiling bounce
		elseif (ball.z + ball.height >= ball.ceilingz)
			bounce = -1
		end
		
		if bounce ~= 0
			ball.momz = bounce * abs(FixedMul(ball.movefactor, BB.BOUNCE_FACTOR))
			S_StartSound(ball, ball.info.activesound)
			BB.ChangeAnimation(ball)
		end
	end
	
	// fan sectors
	if CA.MobjInSectorSpecial(ball, 4, 5)
		BB.DoFan(ball, mobjinfo[MT_FAN].mass, 1)
	end
	
	// buoyancy
	if ball.z <= ball.watertop
	and ball.z + ball.height >= ball.waterbottom
		local flip = P_MobjFlip(ball)
		local buoyancy = FixedMul(BB.BUOYANCY, ball.scale)
		P_SetObjectMomZ(ball, FixedMul(P_GetMobjGravity(ball), -3*FRACUNIT), true)
		
		if flip * ball.momz > buoyancy
			ball.momz = flip * buoyancy
		end
	end
	
	// count down time since last player touch
	if ball.movecount > 0
		ball.movecount = $ - 1
	end
	
	// track previous momz
	ball.movefactor = ball.momz
	
	// set frame
	if ball.threshold > 0
		if P_IsObjectOnGround(ball)
			BB.UpdateAnimationSpeed(ball)
		end
		ball.threshold = $ - ball.extravalue2
	end
	
	if ball.fuse > 0 and ball.fuse < 3*TICRATE
		ball.flags2 = $ ^^ MF2_DONTDRAW
	end
	
	if ball.threshold <= 0
		local frame = ball.frame + 1
		local anim = BB.animations[ball.extravalue1 + 1]
		
		if frame > anim.endFrame
			frame = anim.startFrame
		end
		
		ball.frame = frame
		ball.threshold = 2 * FixedMul(CA.PI, ball.radius) / (anim.endFrame - anim.startFrame)
	end
end, MT_CHRISPY_BEACHBALL)

addHook("MobjFuse", function(ball)
	local mt = ball.spawnpoint
	
	if not mt
		return
	end
	
	local x, y, z = mt.x * FRACUNIT, mt.y * FRACUNIT
	local sector = R_PointInSubsector(x, y).sector
	
	if mt.options & MTF_OBJECTFLIP
		if sector.c_slope
			z = P_GetZAt(sector.c_slope, x, y)
		else
			z = sector.ceilingheight
		end
		z = $ - ball.height - (mt.z * FRACUNIT)
	else
		if sector.f_slope
			z = P_GetZAt(sector.f_slope, x, y)
		else
			z = sector.floorheight
		end
		z = $ + (mt.z * FRACUNIT)
	end
	
	ball.momx = 0
	ball.momy = 0
	ball.momz = 0
	ball.movefactor = 0
	ball.flags2 = $ & ~MF2_DONTDRAW
	BB.ChangeAnimation(ball)
	P_TeleportMove(ball, x, y, z)
	
	return true
end, MT_CHRISPY_BEACHBALL)

addHook("MobjThinker", function(devil)
	if not BB.inLevel
		return
	end
	
	local radius = devil.radius
	searchBlockmap("objects", BB.DustDevilSearch, devil,
		devil.x - radius, devil.x + radius, devil.y - radius, devil.y + radius)
end, MT_DUSTDEVIL)

addHook("MapChange", function()
	BB.inLevel = false
end)

addHook("NetVars", function(net)
	BB.inLevel = net(BB.inLevel)
end)