/*
Bowla Pins
	by Lach

v1
*/

local CA = ChrispyAssets

if CA.BowlaPin
	return
end

local BP = {}

CA.BowlaPin = BP

// SOC

CA.SafeFreeslot("SPR_BPIN", "S_BOWLAPIN", "S_BOWLAPIN_FADE", "MT_BOWLAPIN")

function BP.Fade(mo)
	local val = mo.extravalue2 + 1
	if val > 9
		P_RemoveMobj(mo)
		return
	end
	
	mo.extravalue2 = val
	mo.frame = $ & ~FF_TRANSMASK | (val << FF_TRANSSHIFT)
end

states[S_BOWLAPIN] = {
	sprite = SPR_BPIN,
	tics = -1
}

states[S_BOWLAPIN_FADE] = {
	sprite = SPR_BPIN,
	tics = 1,
	action = BP.Fade,
	nextstate = S_BOWLAPIN_FADE
}

mobjinfo[MT_BOWLAPIN] = {
	--$Name Bowla Pin
	--$Sprite BPINA0
	--$Category Chrispy Pixels
	--$AngleText Lane
	doomednum = 835,
	spawnhealth = 1000,
	spawnstate = S_BOWLAPIN,
	deathstate = S_BOWLAPIN_FADE,
	radius = 8*FRACUNIT,
	height = 64*FRACUNIT,
	flags = MF_SPECIAL|MF_BOUNCE,
	painsound = sfx_bowl
}

// constants

BP.MAX_RECURSION = 32
BP.PIN_FUSE = 3*TICRATE
BP.MF2_SPAWNING = MF2_DONTRESPAWN
BP.MF2_KNOCKED = MF2_SLIDEPUSH
BP.MF2_LANDED = MF2_CLASSICPUSH

// tables

BP.lanes = {}
BP.laneTickers = {}

// functions

function BP.AddPin(pin)
	local i = pin.extravalue1
	if not BP.lanes[i]
		BP.lanes[i] = {}
	end
	table.insert(BP.lanes[i], pin)
end

function BP.RespawnPin(pin)
	local mt = pin.spawnpoint
	
	if not mt
		return
	end
	
	local x, y = mt.x * FRACUNIT, mt.y * FRACUNIT
	local sector = R_PointInSubsector(x, y).sector
	local z, newPin
	
	if mt.options & MTF_OBJECTFLIP
		if sector.f_slope
			z = P_GetZAt(sector.f_slope, x, y)
		else
			z = sector.floorheight
		end
	else
		if sector.c_slope
			z = P_GetZAt(sector.c_slope, x, y)
		else
			z = sector.ceilingheight
		end
	end
	
	P_KillMobj(pin, nil, pin.target)
	
	newPin = P_SpawnMobj(x, y, z, MT_BOWLAPIN)
	newPin.spawnpoint = mt
	mt.mobj = newPin
	
	if mt.options & MTF_OBJECTFLIP
		newPin.flags2 = $ | MF2_OBJECTFLIP
		newPin.eflags = $ | MFE_VERTICALFLIP
	end
	newPin.flags = $ | MF_NOGRAVITY
	newPin.height = 0
	newPin.spriteyscale = 0
	
	BP.MapThingSpawn(newPin, mt)
end

function BP.AddLaneTicker(pin)
	local lane = BP.lanes[pin.extravalue1]
	for _, l in ipairs(BP.laneTickers)
		if lane == l
			return
		end
	end
	table.insert(BP.laneTickers, lane)
end

function BP.PinsInLaneAreActive(lane)
	for i = #lane, 1, -1
		local pin = lane[i]
		
		if not pin.valid
			table.remove(lane, i)
			continue
		end
		
		if pin.fuse == 0
			continue
		end
		
		if pin.momx
		or pin.momy
		or not P_IsObjectOnGround(pin)
			return true
		end
	end
	return false
end

function BP.FindKnockingPlayerMobj(mo)
	local target = mo
	local recursion = 0
	
	while target and recursion < BP.MAX_RECURSION
		if target.player
			return target
		end
		
		recursion = $ + 1
		target = target.target
	end
end

function BP.MapThingSpawn(pin, mt)
	pin.extravalue1 = mt.angle + 1
	BP.AddPin(pin)
end

function BP.MobjThinker(pin)
	// spawn routine
	if pin.flags2 & BP.MF2_SPAWNING
		if P_IsObjectOnGround(pin)
			pin.flags2 = $ & ~BP.MF2_SPAWNING
		elseif pin.flags & MF_NOGRAVITY
			local scale = pin.scale
			local goalHeight = FixedMul(pin.info.height, scale)
			
			if pin.height < goalHeight
				local newHeight = pin.height + 4*scale
				if newHeight >= goalHeight
					newHeight = goalHeight
					pin.flags = $ & ~MF_NOGRAVITY
				end
				pin.height = newHeight
				pin.spriteyscale = FixedDiv(newHeight, goalHeight)
			end
		end
		return
	end
	
	// no thinking if the pin hasn't been knocked
	if pin.flags2 & (BP.MF2_KNOCKED|BP.MF2_LANDED) ~= BP.MF2_KNOCKED
		return
	end
	
	local info = pin.info
	local scale = pin.scale
	
	if P_IsObjectOnGround(pin)
		if pin.rollangle < 0
			pin.rollangle = -ANGLE_90
		else
			pin.rollangle = ANGLE_90
		end
		pin.spriteyoffset = $ - (info.height >> 1) + info.radius
		pin.flags2 = $ | BP.MF2_LANDED
	else
		pin.rollangle = $ + pin.movedir
	end
	
	pin.flags = $ | MF_NOCLIP
	pin.radius = FixedMul(scale, info.radius + FixedMul(abs(sin(pin.rollangle)), (info.height >> 1) - info.radius))
	pin.flags = $ & ~MF_NOCLIP
end

function BP.MobjCollide(pin, mo)
	if pin.flags2 & (BP.MF2_KNOCKED | BP.MF2_SPAWNING)
	or pin.health <= 0
	or not (mo.momx or mo.momy)
		return
	end
	
	local playerMo = BP.FindKnockingPlayerMobj(mo)
	
	if not playerMo
		return
	end
	
	local info = pin.info
	local speed = FixedHypot(mo.momx, mo.momy)
	local momAngle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	local physAngle = R_PointToAngle2(mo.x, mo.y, pin.x, pin.y)
	
	P_InstaThrust(mo, 2*momAngle - physAngle, speed)
	P_InstaThrust(pin, physAngle, speed)
	P_SetObjectMomZ(pin, 4*FRACUNIT)
	pin.z = $ + P_MobjFlip(pin)
	S_StartSound(pin, pin.info.painsound)
	
	pin.fuse = BP.PIN_FUSE
	pin.movedir = (2 * P_RandomKey(2) - 1) * FixedAngle(speed)
	pin.flags2 = $ | BP.MF2_KNOCKED
	pin.target = playerMo
	
	BP.AddLaneTicker(pin)
end

function BP.Ticker()
	for i = #BP.laneTickers, 1, -1
		local lane = BP.laneTickers[i]
		if BP.PinsInLaneAreActive(lane)
			continue
		end
		
		local numPins = 0
		
		table.remove(BP.laneTickers, i)
		
		for j = #lane, 1, -1
			local pin = lane[j]
			if pin.valid
				if pin.flags2 & BP.MF2_KNOCKED
					numPins = $ + 1
				end
				BP.RespawnPin(pin)
			end
			table.remove(lane, j)
		end
	end
end

// hooks

addHook("MobjSpawn", function(pin) pin.flags2 = $ | BP.MF2_SPAWNING end, MT_BOWLAPIN)
addHook("MapThingSpawn", BP.MapThingSpawn, MT_BOWLAPIN)
addHook("MobjCollide", BP.MobjCollide, MT_BOWLAPIN)
addHook("MobjThinker", BP.MobjThinker, MT_BOWLAPIN)
addHook("TouchSpecial", do return true end, MT_BOWLAPIN)
addHook("MobjFuse", do return true end, MT_BOWLAPIN)
addHook("MobjMoveBlocked", function(pin) pin.momx, pin.momy = $1/2, $2/2 end, MT_BOWLAPIN)

addHook("ThinkFrame", BP.Ticker)
addHook("MapChange", function()
	BP.lanes = {}
	BP.laneTickers = {}
end)
addHook("NetVars", function(net)
	BP.lanes = net(BP.lanes)
	BP.laneTickers = net(BP.laneTickers)
end)

// make Rollout Rocks target their rider
addHook("MobjThinker", function(rock)
	local tracer = rock.tracer
	if tracer ~= nil
		rock.target = tracer
	end
end, MT_ROLLOUTROCK)