/*
Chrispy Assets
	by Lach

v1
*/

local SCRIPT_VERSION = 1
local CA = ChrispyAssets

if CA == nil
	CA = {}
	rawset(_G, "ChrispyAssets", CA)
elseif CA.version >= VERSION
	return
end

CA.version = SCRIPT_VERSION

// shortcuts
CA.MF_INTANGIBLE = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING

// constants
CA.ORIG_FRICTION = 29*FRACUNIT/32
CA.PI = 355*FRACUNIT/113 // Zu Chongzi approximation

// To prevent duplicate freeslots
local function CheckSlot(item) // this function deliberately errors when a freeslot does not exist
	if _G[item] == nil // this will error by itself for states and objects
		error() // this forces an error for sprites, which do actually return nil for some reason
	end
end

function CA.SafeFreeslot(...)
	local item = ...
	if item == nil
		return
	end
	
	if pcall(CheckSlot, item)
		print("\131NOTICE:\128 " .. item .. " was not allocated, as it already exists.")
	else
		freeslot(item)
	end
	
	CA.SafeFreeslot(select(2, ...))
end

function CA.MobjsIntersect(m1, m2, doZChecks)
	local r1 = m1.radius
	local r2 = m2.radius
	
	if doZChecks
		if m1.z > m2.z + m2.height + FRACUNIT
		or m2.z > m1.z + m1.height + FRACUNIT
			return false
		end
	end
	
	return m1.x + r1 > m2.x - r2
	and m1.x - r1 < m2.x + r2
	and m1.y + r1 > m2.y - r2
	and m1.y - r1 < m2.y + r2
end

function CA.ButteredSlope(mo)
	local slope = mo.standingslope
	
	if not slope
	or slope.flags & SL_NOPHYSICS
	or mo.flags & (MF_NOCLIPHEIGHT|MF_NOGRAVITY)
		return false
	end
	
	local player = mo.player
	
	if player
		local delta = abs(slope.zdelta)
		if zdelta < FRACUNIT/4 and not (player.pflags & PF_SPINNING)
		or zdelta < FRACUNIT/2 and not (player.rmomx or player.rmomy)
			return false
		end
	end
	
	local momx, momy = mo.momx, mo.momy
	local flip = P_MobjFlip(mo)
	local thrust = sin(slope.zangle) * 3 / 2 * -flip
	
	if player and player.pflags & PF_SPINNING
		local mult = 0
		if (momx or momy)
			local angle = R_PointToAngle2(0, 0, momx, momy) - slope.xydirection
			
			if flip * slope.zdelta < 0
				angle = $ + ANGLE_180
			end
			
			mult = cos(angle)
		end
		
		thrust = FixedMul($, FRACUNIT*2/3 + mult/8)
	end
	
	if momx or momy
		thrust = FixedMul($, FRACUNIT + P_AproxDistance(momx, momy)/16)
	end
	
	thrust = FixedMul($, abs(P_GetMobjGravity(mo)))
	thrust = FixedMul($, FixedDiv(mo.friction, CA.ORIG_FRICTION))
	
	P_Thrust(mo, slope.xydirection, thrust)
	return true
end

function CA.GetFOFHeightsAt(fof, x, y)
	local top
	local bottom
	
	if fof.t_slope
		top = P_GetZAt(fof.t_slope, x, y)
	else
		top = fof.topheight
	end
	
	if fof.b_slope
		bottom = P_GetZAt(fof.b_slope, x, y)
	else
		bottom = fof.bottomheight
	end
	
	return top, bottom
end

function CA.MobjInSectorSpecial(mo, section, number)
	local sector = mo.subsector.sector
	
	if GetSecSpecial(sector.special, section) == number
		return sector
	end
	
	for fof in sector.ffloors()
		sector = fof.sector
		if GetSecSpecial(sector.special, section) ~= number
		or not (fof.flags & FF_EXISTS)
			continue
		end
		
		local top, bottom = CA.GetFOFHeightsAt(fof, mo.x, mo.y)
		
		if mo.z > top
		or mo.z + mo.height < bottom
			continue
		end
		
		return sector
	end
end

if not CA.HAVE_CAMERA
	CA.HAVE_CAMERA = true
	if camera
		CA.camera = camera
		CA.camera2 = camera2
	else
		// camera access in 2.2.9 and below
		hud.add(function(v, player, cam)
			if splitscreen and player == players[1]
				CA.camera2 = cam
			else
				CA.camera = cam
			end
		end)
	end
end